"""Auto-generated constants from static game data."""

ELIXIR_TROOP_ORDER = [
    'Barbarian',
    'Archer',
    'Goblin',
    'Giant',
    'Wall Breaker',
    'Balloon',
    'Wizard',
    'Healer',
    'Dragon',
    'P.E.K.K.A',
    'Baby Dragon',
    'Miner',
    'Yeti',
    'Electro Dragon',
    'Dragon Rider',
    'Electro Titan',
    'Root Rider',
    'Thrower',
    'Meteor Golem',
]

DARK_ELIXIR_TROOP_ORDER = [
    'Minion',
    'Hog Rider',
    'Valkyrie',
    'Golem',
    'Witch',
    'Lava Hound',
    'Bowler',
    'Ice Golem',
    'Headhunter',
    'Apprentice Warden',
    'Druid',
    'Furnace',
]

HV_TROOP_ORDER = ELIXIR_TROOP_ORDER + DARK_ELIXIR_TROOP_ORDER

SIEGE_MACHINE_ORDER = [
    'Wall Wrecker',
    'Battle Blimp',
    'Stone Slammer',
    'Siege Barracks',
    'Log Launcher',
    'Flame Flinger',
    'Battle Drill',
    'Troop Launcher',
]

SUPER_TROOP_ORDER = [
    'Super Barbarian',
    'Super Archer',
    'Super Wall Breaker',
    'Super Giant',
    'Sneaky Goblin',
    'Super Miner',
    'Rocket Balloon',
    'Inferno Dragon',
    'Super Valkyrie',
    'Super Witch',
    'Ice Hound',
    'Super Bowler',
    'Super Dragon',
    'Super Wizard',
    'Super Minion',
    'Super Hog Rider',
    'Super Yeti',
]

HOME_TROOP_ORDER = HV_TROOP_ORDER + SIEGE_MACHINE_ORDER

SEASONAL_TROOP_ORDER = [
    'Ice Wizard',
    'Battle Ram',
    'Royal Ghost',
    'Pumpkin Barbarian',
    'Giant Skeleton',
    'Skeleton Barrel',
    'M.E.C.H.A',
    'Party Wizard',
    'Ram Rider',
    'Barcher',
    'Witch Golem',
    'Hog Wizard',
    'Lavaloon',
    'C.O.O.K.I.E',
    'Firecracker',
    'Azure Dragon',
    'Barbarian Kicker',
    'Giant Thrower',
    'Broom Witch',
    'Ice Minion',
    'Debt Collector',
    'Snake Barrel',
    'Giant Giant',
    'K.A.N.E',
    'The Disarmer',
    'YEETer',
    'YEETer',
    'The Disarmer',
    'Meteor Golem',
]

BUILDER_TROOPS_ORDER = [
    'Raged Barbarian',
    'Sneaky Archer',
    'Beta Minion',
    'Boxer Giant',
    'Bomber',
    'Power P.E.K.K.A',
    'Cannon Cart',
    'Drop Ship',
    'Baby Dragon',
    'Night Witch',
    'Hog Glider',
    'Electrofire Wizard',
]

ELIXIR_SPELL_ORDER = [
    'Lightning Spell',
    'Healing Spell',
    'Rage Spell',
    'Jump Spell',
    'Freeze Spell',
    'Clone Spell',
    'Invisibility Spell',
    'Recall Spell',
    'Revive Spell',
    'Totem Spell',
]

DARK_ELIXIR_SPELL_ORDER = [
    'Poison Spell',
    'Earthquake Spell',
    'Haste Spell',
    'Skeleton Spell',
    'Bat Spell',
    'Overgrowth Spell',
    'Ice Block Spell',
]

SEASONAL_SPELL_ORDER = [
    "Santa's Surprise",
    'Bag of Frostmites',
]

SPELL_ORDER = ELIXIR_SPELL_ORDER + DARK_ELIXIR_SPELL_ORDER

HOME_BASE_HERO_ORDER = [
    'Barbarian King',
    'Archer Queen',
    'Minion Prince',
    'Grand Warden',
    'Royal Champion',
]

BUILDER_BASE_HERO_ORDER = [
    'Battle Machine',
    'Battle Copter',
]

HERO_ORDER = HOME_BASE_HERO_ORDER + BUILDER_BASE_HERO_ORDER

PETS_ORDER = [
    'L.A.S.S.I',
    'Mighty Yak',
    'Electro Owl',
    'Unicorn',
    'Phoenix',
    'Poison Lizard',
    'Diggy',
    'Frosty',
    'Spirit Fox',
    'Angry Jelly',
    'Sneezy',
]

EQUIPMENT = [
    'Barbarian Puppet',
    'Rage Vial',
    'Archer Puppet',
    'Invisibility Vial',
    'Eternal Tome',
    'Life Gem',
    'Seeking Shield',
    'Royal Gem',
    'Earthquake Boots',
    'Hog Rider Puppet',
    'Giant Gauntlet',
    'Vampstache',
    'Haste Vial',
    'Rocket Spear',
    'Spiky Ball',
    'Frozen Arrow',
    'Giant Arrow',
    'Heroic Torch',
    'Healer Puppet',
    'Fireball',
    'Rage Gem',
    'Snake Bracelet',
    'Healing Tome',
    'Dark Crown',
    'Magic Mirror',
    'Electro Boots',
    'Lavaloon Puppet',
    'Henchmen Puppet',
    'Dark Orb',
    'Metal Pants',
    'Noble Iron',
    'Action Figure',
    'Meteor Staff',
    'Frost Flake',
    'Stick Horse',
]

HV_BUILDINGS = [
    'Army Camp',
    'Town Hall',
    'Elixir Collector',
    'Elixir Storage',
    'Gold Mine',
    'Gold Storage',
    'Barracks',
    'Laboratory',
    'Cannon',
    'Archer Tower',
    'Wall',
    'Wizard Tower',
    'Air Defense',
    'Mortar',
    'Clan Castle',
    "Builder's Hut",
    'Hidden Tesla',
    'Spell Factory',
    'X-Bow',
    'Dark Elixir Drill',
    'Dark Elixir Storage',
    'Dark Barracks',
    'Inferno Tower',
    'Air Sweeper',
    'Dark Spell Factory',
    'Eagle Artillery',
    'Bomb Tower',
    'Workshop',
    "B.O.B's Hut",
    'Scattershot',
    'Pet House',
    'Blacksmith',
    'Hero Hall',
    'Spell Tower',
    'Monolith',
    'Multi-Gear Tower',
    'Sour Elixir Cauldron',
    'Multi-Archer Tower',
    'Ricochet Cannon',
    'Revenge Tower',
    'Firespitter',
    'Helper Hut',
    'Crafting Station',
    'Super Wizard Tower',
]

ACHIEVEMENT_ORDER = [
    'Keep Your Account Safe!',
    'Bigger & Better',
    'Discover New Troops',
    'Bigger Coffers',
    'Gold Grab',
    'Elixir Escapade',
    'Heroic Heist',
    'Well Seasoned',
    'Nice and Tidy',
    'Empire Builder',
    'Clan War Wealth',
    'Friend in Need',
    'Sharing is caring',
    'Siege Sharer',
    'War Hero',
    'War League Legend',
    'Games Champion',
    'Unbreakable',
    'Sweet Victory!',
    'Conqueror',
    'League All-Star',
    'League Follower',
    'League Enthusiast',
    'League Superfan',
    'League Fanatic',
    'League Master',
    'Humiliator',
    'Not So Easy This Time',
    'Union Buster',
    'Bust This!',
    'Wall Buster',
    'Mortar Mauler',
    'X-Bow Exterminator',
    'Firefighter',
    'Anti-Artillery',
    'Shattered and Scattered',
    'Counterspell',
    'Monolith Masher',
    'Multi-Archer Tower Terminator',
    'Ricochet Cannon Crusher',
    'Firespitter Finisher',
    'Multi-Gear Tower Trampler',
    'Crafter’s Nightmare',
    'Get those Goblins!',
    'Supercharger',
    'Crafting Connoisseur',
    'Get those other Goblins!',
    'Get even more Goblins!',
    'Dragon Slayer',
    'Ungrateful Child',
    'Superb Work',
    'Master Engineering',
    'Hidden Treasures',
    'High Gear',
    'Next Generation Model',
    'Un-Build It',
    'Champion Builder',
    'Aggressive Capitalism',
    'Most Valuable Clanmate',
]

